/*!@file Apps/BorderWatch/BorderWatchGui.C Simple GUI for BorderWatch results */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2005   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti
// $HeadURL: svn://isvn.usc.edu/software/invt/trunk/saliency/src/Apps/BorderWatch/BorderWatchGui.C $
// $Id: BorderWatchGui.C 15310 2012-06-01 02:29:24Z itti $
//

#include "QtUtil/Util.H" // for argv2qt()
#include "Apps/BorderWatch/BorderWatchQt.qt.H"
#include "Apps/BorderWatch/BorderWatchData.H"
#include "Component/ModelManager.H"
#include "Component/ModelOptionDef.H"
#include "Image/Image.H"
#include "Media/FrameSeries.H"
#include "Raster/GenericFrame.H"

#include <QtGui/QApplication>
#include <fstream>

const ModelOptionCateg MOC_BORDERWATCH = {
  MOC_SORTPRI_3, "BorderWatch-Related Options" };

const ModelOptionDef OPT_LogFile =
  { MODOPT_ARG_STRING, "LogFile", &MOC_BORDERWATCH, OPTEXP_CORE,
    "File name of the log file generated by BorderWatch application",
    "logfile", '\0', "<file.log>", "" };

//! BorderWatch GUI, use with --in=movie.asf --logfile=file.log
int main(int argc, const char **argv)
{
  // Instantiate a ModelManager:
  ModelManager manager("Attention Model");

  nub::ref<InputFrameSeries> ifs(new InputFrameSeries(manager));
  manager.addSubComponent(ifs);

  OModelParam<std::string> optLogFile(&OPT_LogFile, &manager);

  manager.exportOptions(MC_RECURSE);

  // Parse command-line:
  if (manager.parseCommandLine(argc, argv, "", 0, 0) == false) return 1;

  if (optLogFile.getVal().empty()) LFATAL("You need to provide a log file with --logfile=<file.log>");

  // let's get all our ModelComponent instances started:
  manager.start();

  // Load up the log file:
  LINFO("Loading logfile data from %s", optLogFile.getVal().c_str());
  std::vector<BorderWatchData> logdata;
  std::ifstream inlog(optLogFile.getVal().c_str());
  if (inlog.is_open() == false) LFATAL("Cannot open %s for reading", optLogFile.getVal().c_str());
  BorderWatchData d;
  while (true) { inlog >> d; if (inlog.eof()) break; else logdata.push_back(d); }
  inlog.close();
  LINFO("     %" ZU " log entries.", logdata.size());

  // Load up the movie frames:
  LINFO("Loading movie frames...");
  std::vector<GenericFrame> frames;
  while (true) {
    // grab the images:
    ifs->updateNext();
    GenericFrame input = ifs->readFrame(); if (input.initialized() == false) break;
    frames.push_back(input);
    if ((ifs->frame() % 5000) == 0) LINFO("     frame %06u", ifs->frame());
  }
  LINFO("     %" ZU " frames.", frames.size());

  // we are done with our ModelComponents:
  manager.stop();

  // create a QApplication:
  LINFO("Starting GUI...");
  int qtargc = 1; const char* qtargv[1] = { "BorderWatchGUI" };
  QApplication a(qtargc, argv2qt(qtargc, qtargv));

  // and a widget:
  BorderWatchQt b(frames, logdata);
  b.setWindowTitle("iLab USC -- BorderWatch Review GUI");
  b.show();

  // main loop for QApplication:
  const int ret = a.exec();

  // cleanup and exit:
  return ret;
}
