#!/usr/bin/perl -w

# $Id: test_ezvision_blackbox.pl 14345 2010-12-27 17:38:55Z jshen $

# This Perl script is a "black box" test driver. The idea is that we
# run the executable with different sets of command-line options, and
# compare the output files with reference files that are stored in the
# ref/ subdirectory. See blackbox.pm for implementation details.

use strict;

use blackbox;
use invt_config;

# here is the key to the test suite: the database of command-line
# options along with the expected output files for those option sets

# note that we're doing some implicit tests for proper parsing of
# --in/--out/--io arguments here, trying different combinations of
# things (e.g., try --in/--out vs. --io, try with and without an
# explicit 'raster:' prefix)

my @tests =
    (
     {
	 name  => "ez-trajectory--1",
	 args  => ['-TZ', '--boring-delay=max', '--nodisplay-eye',
		   '--nodisplay-eye-traj', '--textlog=test.txt',
		   '--output-frames=0-5@100',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm',
		   'T000004.pnm', 'T000005.pnm'],
     },

     {
	 name  => "ez-vc-type-entropy--1",
	 args  => ['-TZ', '--boring-delay=MAX', '--nodisplay-eye',
		   '--nodisplay-eye-traj', '--textlog=test.txt',
		   '--output-frames=0-5@100', '--vc-type=Entrop', '--maxnorm-type=Ignore ',
		   '--in=testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm',
		   'T000004.pnm', 'T000005.pnm'],
     },

     {
	 name  => "ez-vc-type-variance--1",
	 args  => ['-TZ', '--boring-delay=forever', '--nodisplay-eye',
		   '--nodisplay-eye-traj', '--textlog=test.txt',
		   '--output-frames=0-5@100', '--vc-type=Variance', '--maxnorm-type=Ignore ',
		   '--in=raster:testpic001.pnm', '-+', '--out=pnm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm',
		   'T000004.pnm', 'T000005.pnm'],
     },

     {
	 name  => "ez-raw-maps--1",
	 args  => ['-Z', '--textlog=test.txt', '--nodisplay-eye',
		   '--nodisplay-eye-traj', '--too-many-shifts=1',
		   '--output-frames=@EVENT', '--save-raw-maps',
		   '--io=raster:testpic001.pnm', '-+'],
	 files => ['test.txt',
		   'testpic001.pnm-SRby-0-000000.pnm', 'testpic001.pnm-SRintensity-6-000000.pnm',
		   'testpic001.pnm-SRby-1-000000.pnm', 'testpic001.pnm-SRintensity-7-000000.pnm',
		   'testpic001.pnm-SRby-2-000000.pnm', 'testpic001.pnm-SRintensity-8-000000.pnm',
		   'testpic001.pnm-SRby-3-000000.pnm', 'testpic001.pnm-SRori_0-0-000000.pnm',
		   'testpic001.pnm-SRby-4-000000.pnm', 'testpic001.pnm-SRori_0-1-000000.pnm',
		   'testpic001.pnm-SRby-5-000000.pnm', 'testpic001.pnm-SRori_0-2-000000.pnm',
		   'testpic001.pnm-SRby-6-000000.pnm', 'testpic001.pnm-SRori_0-3-000000.pnm',
		   'testpic001.pnm-SRby-7-000000.pnm', 'testpic001.pnm-SRori_0-4-000000.pnm',
		   'testpic001.pnm-SRby-8-000000.pnm', 'testpic001.pnm-SRori_0-5-000000.pnm',
		   'testpic001.pnm-SRdir_0-0-000000.pnm', 'testpic001.pnm-SRori_0-6-000000.pnm',
		   'testpic001.pnm-SRdir_0-1-000000.pnm', 'testpic001.pnm-SRori_0-7-000000.pnm',
		   'testpic001.pnm-SRdir_0-2-000000.pnm', 'testpic001.pnm-SRori_0-8-000000.pnm',
		   'testpic001.pnm-SRdir_0-3-000000.pnm', 'testpic001.pnm-SRori_1-0-000000.pnm',
		   'testpic001.pnm-SRdir_0-4-000000.pnm', 'testpic001.pnm-SRori_1-1-000000.pnm',
		   'testpic001.pnm-SRdir_0-5-000000.pnm', 'testpic001.pnm-SRori_1-2-000000.pnm',
		   'testpic001.pnm-SRdir_0-6-000000.pnm', 'testpic001.pnm-SRori_1-3-000000.pnm',
		   'testpic001.pnm-SRdir_0-7-000000.pnm', 'testpic001.pnm-SRori_1-4-000000.pnm',
		   'testpic001.pnm-SRdir_0-8-000000.pnm', 'testpic001.pnm-SRori_1-5-000000.pnm',
		   'testpic001.pnm-SRdir_1-0-000000.pnm', 'testpic001.pnm-SRori_1-6-000000.pnm',
		   'testpic001.pnm-SRdir_1-1-000000.pnm', 'testpic001.pnm-SRori_1-7-000000.pnm',
		   'testpic001.pnm-SRdir_1-2-000000.pnm', 'testpic001.pnm-SRori_1-8-000000.pnm',
		   'testpic001.pnm-SRdir_1-3-000000.pnm', 'testpic001.pnm-SRori_2-0-000000.pnm',
		   'testpic001.pnm-SRdir_1-4-000000.pnm', 'testpic001.pnm-SRori_2-1-000000.pnm',
		   'testpic001.pnm-SRdir_1-5-000000.pnm', 'testpic001.pnm-SRori_2-2-000000.pnm',
		   'testpic001.pnm-SRdir_1-6-000000.pnm', 'testpic001.pnm-SRori_2-3-000000.pnm',
		   'testpic001.pnm-SRdir_1-7-000000.pnm', 'testpic001.pnm-SRori_2-4-000000.pnm',
		   'testpic001.pnm-SRdir_1-8-000000.pnm', 'testpic001.pnm-SRori_2-5-000000.pnm',
		   'testpic001.pnm-SRdir_2-0-000000.pnm', 'testpic001.pnm-SRori_2-6-000000.pnm',
		   'testpic001.pnm-SRdir_2-1-000000.pnm', 'testpic001.pnm-SRori_2-7-000000.pnm',
		   'testpic001.pnm-SRdir_2-2-000000.pnm', 'testpic001.pnm-SRori_2-8-000000.pnm',
		   'testpic001.pnm-SRdir_2-3-000000.pnm', 'testpic001.pnm-SRori_3-0-000000.pnm',
		   'testpic001.pnm-SRdir_2-4-000000.pnm', 'testpic001.pnm-SRori_3-1-000000.pnm',
		   'testpic001.pnm-SRdir_2-5-000000.pnm', 'testpic001.pnm-SRori_3-2-000000.pnm',
		   'testpic001.pnm-SRdir_2-6-000000.pnm', 'testpic001.pnm-SRori_3-3-000000.pnm',
		   'testpic001.pnm-SRdir_2-7-000000.pnm', 'testpic001.pnm-SRori_3-4-000000.pnm',
		   'testpic001.pnm-SRdir_2-8-000000.pnm', 'testpic001.pnm-SRori_3-5-000000.pnm',
		   'testpic001.pnm-SRdir_3-0-000000.pnm', 'testpic001.pnm-SRori_3-6-000000.pnm',
		   'testpic001.pnm-SRdir_3-1-000000.pnm', 'testpic001.pnm-SRori_3-7-000000.pnm',
		   'testpic001.pnm-SRdir_3-2-000000.pnm', 'testpic001.pnm-SRori_3-8-000000.pnm',
		   'testpic001.pnm-SRdir_3-3-000000.pnm', 'testpic001.pnm-SRrg-0-000000.pnm',
		   'testpic001.pnm-SRdir_3-4-000000.pnm', 'testpic001.pnm-SRrg-1-000000.pnm',
		   'testpic001.pnm-SRdir_3-5-000000.pnm', 'testpic001.pnm-SRrg-2-000000.pnm',
		   'testpic001.pnm-SRdir_3-6-000000.pnm', 'testpic001.pnm-SRrg-3-000000.pnm',
		   'testpic001.pnm-SRdir_3-7-000000.pnm', 'testpic001.pnm-SRrg-4-000000.pnm',
		   'testpic001.pnm-SRdir_3-8-000000.pnm', 'testpic001.pnm-SRrg-5-000000.pnm',
		   'testpic001.pnm-SRintensity-0-000000.pnm', 'testpic001.pnm-SRrg-6-000000.pnm',
		   'testpic001.pnm-SRintensity-1-000000.pnm', 'testpic001.pnm-SRrg-7-000000.pnm',
		   'testpic001.pnm-SRintensity-2-000000.pnm', 'testpic001.pnm-SRrg-8-000000.pnm',
		   'testpic001.pnm-SRintensity-3-000000.pnm',
		   'testpic001.pnm-SRintensity-4-000000.pnm',
		   'testpic001.pnm-SRintensity-5-000000.pnm'],
     },

     {
	 name  => "ez-feature-maps--1",
	 args  => ['-Z', '--textlog=test.txt', '--too-many-shifts=1',
		   '--output-frames=@EVENT', '--save-feature-maps',
		   '--in=testpic001.pnm', '-+', '--out=testpic001.pnm'],
	 files => ['test.txt',
		   'testpic001.pnm-SFby-2-5-000000.pnm', 'testpic001.pnm-SFintensity-4-7-000000.pnm',
		   'testpic001.pnm-SFby-2-6-000000.pnm', 'testpic001.pnm-SFintensity-4-8-000000.pnm',
		   'testpic001.pnm-SFby-3-6-000000.pnm', 'testpic001.pnm-SFori_0-2-5-000000.pnm',
		   'testpic001.pnm-SFby-3-7-000000.pnm', 'testpic001.pnm-SFori_0-2-6-000000.pnm',
		   'testpic001.pnm-SFby-4-7-000000.pnm', 'testpic001.pnm-SFori_0-3-6-000000.pnm',
		   'testpic001.pnm-SFby-4-8-000000.pnm', 'testpic001.pnm-SFori_0-3-7-000000.pnm',
		   'testpic001.pnm-SFdir_0-2-5-000000.pnm', 'testpic001.pnm-SFori_0-4-7-000000.pnm',
		   'testpic001.pnm-SFdir_0-2-6-000000.pnm', 'testpic001.pnm-SFori_0-4-8-000000.pnm',
		   'testpic001.pnm-SFdir_0-3-6-000000.pnm', 'testpic001.pnm-SFori_1-2-5-000000.pnm',
		   'testpic001.pnm-SFdir_0-3-7-000000.pnm', 'testpic001.pnm-SFori_1-2-6-000000.pnm',
		   'testpic001.pnm-SFdir_0-4-7-000000.pnm', 'testpic001.pnm-SFori_1-3-6-000000.pnm',
		   'testpic001.pnm-SFdir_0-4-8-000000.pnm', 'testpic001.pnm-SFori_1-3-7-000000.pnm',
		   'testpic001.pnm-SFdir_1-2-5-000000.pnm', 'testpic001.pnm-SFori_1-4-7-000000.pnm',
		   'testpic001.pnm-SFdir_1-2-6-000000.pnm', 'testpic001.pnm-SFori_1-4-8-000000.pnm',
		   'testpic001.pnm-SFdir_1-3-6-000000.pnm', 'testpic001.pnm-SFori_2-2-5-000000.pnm',
		   'testpic001.pnm-SFdir_1-3-7-000000.pnm', 'testpic001.pnm-SFori_2-2-6-000000.pnm',
		   'testpic001.pnm-SFdir_1-4-7-000000.pnm', 'testpic001.pnm-SFori_2-3-6-000000.pnm',
		   'testpic001.pnm-SFdir_1-4-8-000000.pnm', 'testpic001.pnm-SFori_2-3-7-000000.pnm',
		   'testpic001.pnm-SFdir_2-2-5-000000.pnm', 'testpic001.pnm-SFori_2-4-7-000000.pnm',
		   'testpic001.pnm-SFdir_2-2-6-000000.pnm', 'testpic001.pnm-SFori_2-4-8-000000.pnm',
		   'testpic001.pnm-SFdir_2-3-6-000000.pnm', 'testpic001.pnm-SFori_3-2-5-000000.pnm',
		   'testpic001.pnm-SFdir_2-3-7-000000.pnm', 'testpic001.pnm-SFori_3-2-6-000000.pnm',
		   'testpic001.pnm-SFdir_2-4-7-000000.pnm', 'testpic001.pnm-SFori_3-3-6-000000.pnm',
		   'testpic001.pnm-SFdir_2-4-8-000000.pnm', 'testpic001.pnm-SFori_3-3-7-000000.pnm',
		   'testpic001.pnm-SFdir_3-2-5-000000.pnm', 'testpic001.pnm-SFori_3-4-7-000000.pnm',
		   'testpic001.pnm-SFdir_3-2-6-000000.pnm', 'testpic001.pnm-SFori_3-4-8-000000.pnm',
		   'testpic001.pnm-SFdir_3-3-6-000000.pnm', 'testpic001.pnm-SFrg-2-5-000000.pnm',
		   'testpic001.pnm-SFdir_3-3-7-000000.pnm', 'testpic001.pnm-SFrg-2-6-000000.pnm',
		   'testpic001.pnm-SFdir_3-4-7-000000.pnm', 'testpic001.pnm-SFrg-3-6-000000.pnm',
		   'testpic001.pnm-SFdir_3-4-8-000000.pnm', 'testpic001.pnm-SFrg-3-7-000000.pnm',
		   'testpic001.pnm-SFintensity-2-5-000000.pnm', 'testpic001.pnm-SFrg-4-7-000000.pnm',
		   'testpic001.pnm-SFintensity-2-6-000000.pnm', 'testpic001.pnm-SFrg-4-8-000000.pnm',
		   'testpic001.pnm-SFintensity-3-6-000000.pnm',
		   'testpic001.pnm-SFintensity-3-7-000000.pnm'],
     },

     {
	 name  => "ez-conspicuity-maps--1",
	 args  => ['-Z', '--textlog=test.txt', '--too-many-shifts=1',
		   '--output-frames=@EVENT', '--save-conspic-maps',
		   '--in=raster:testpic001.pnm', '-+',
		   '--out=raster:testpic001.pnm'],
	 files => ['test.txt',
		   'testpic001.pnm-COcolor-000000.pnm',
		   'testpic001.pnm-COorientation-000000.pnm',
		   'testpic001.pnm-COmotion-000000.pnm'],
     },

     {
	 name  => "ez-conspicuity-maps--2",
	 args  => ['-Z', '--textlog=test.txt', '--too-many-shifts=1',
		   '--output-frames=@EVENT', '--save-conspic-maps',
		   '--maxnorm-type=Maxnorm', '--in=raster:testpic001.pnm',
		   '-+', '--out=raster:testpic001.pnm'],
	 files => ['test.txt',
		   'testpic001.pnm-COcolor-000000.pnm',
		   'testpic001.pnm-COorientation-000000.pnm',
		   'testpic001.pnm-COmotion-000000.pnm'],
     },

     {
	 name  => "ez-saliency-map--1",
	 args  => ['-Z', '--textlog=test.txt', '--too-many-shifts=1',
		   '--output-frames=@EVENT', '--save-salmap',
		   '--in=raster:testpic001.pnm', '-+',
		   '--out=raster:testpic001.pnm'],
	 files => ['test.txt', 'testpic001.pnm-SM000000.pfm'],
     },

     {
	 name  => "ez-saliency-map--2",
	 args  => ['-Z', '--textlog=test.txt', '--too-many-shifts=1',
		   '--output-frames=@EVENT', '--save-salmap',
		   '--maxnorm-type=Maxnorm', '--in=raster:testpic001.pnm',
		   '-+', '--out=raster:testpic001.pnm'],
	 files => ['test.txt', 'testpic001.pnm-SM000000.pfm'],
     },

     {
	 name  => "ez-saliency-map--3",
	 args  => ['-Z', '--textlog=test.txt', '--output-frames=0-4@350',
		   '--save-salmap', '--in=raster:testpic001.pnm',
		   '-+', '--out=raster:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm', 'SM000004.pfm'],
     },

     {
	 name  => "ez-saliency-map--4",
	 args  => ['-Z', '--textlog=test.txt', '--output-frames=0-4@350',
		   '--save-salmap', '--maxnorm-type=Maxnorm',
		   '--in=raster:testpic001.pnm', '-+', '--out=pnm:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm', 'SM000004.pfm'],
     },

     {
	 name  => "ez-foa--1",
	 args  => ['-ZT', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--nodisplay-eye', '--nodisplay-eye-traj',
		   '--textlog=test.txt', '--output-frames=0-4@250',
		   '--crop-foa=64x64',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-smmod--1",
	 args  => ['-ZT', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--nodisplay-eye', '--nodisplay-eye-traj',
		   '--textlog=test.txt', '--output-frames=0-4@250',
		   '--display-smmod',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-display--1",
	 args  => ['-Z', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--nodisplay-eye', '--nodisplay-eye-traj',
		   '--textlog=test.txt', '--output-frames=0-4@250',
		   '--mega-combo',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-display--2",
	 args  => ['-Z', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--nodisplay-eye', '--nodisplay-eye-traj',
		   '--textlog=test.txt', '--output-frames=0-4@250',
		   '--mega-combo', '--nodisplay-interp-maps',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-display--3",
	 args  => ['-Z', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--nodisplay-eye', '--nodisplay-eye-traj',
		   '--textlog=test.txt', '--output-frames=0-4@250',
		   '--warp-salmap-3d',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-display--4",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--nodisplay-eye', '--nodisplay-eye-traj',
		   '--textlog=test.txt', '--output-frames=0-4@250',
		   '--vc-chans=CIOLTXEK', '--display-foanum',
		   '--nodisplay-interp-maps', '--maxnorm-type=FancyWeak',
		   '--nomega-combo-topcm',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-svcompr--1",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa',
		   '--sv-type=Compress', '--nodisplay-interp-maps', '--maxnorm-type=Maxnorm',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz', '--wta-type=None' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-svcompr--2",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa', '--display-patch',
		   '--sv-type=Compress', '--nodisplay-interp-maps', '--maxnorm-type=Maxnorm',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz', '--num-foveas=3',
		   '--out=png', '--output-frames=@6Hz', '--wta-type=None' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-svcompr--3",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa',
		   '--sv-type=Compress', '--nodisplay-interp-maps', '--maxnorm-type=Maxnorm',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz', '--wta-type=None',
		   '--num-foveas=3', '--trm-type=KillStatic', '--use-trmmax' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-intvcx--1",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa',
		   '--vc-type=Int', '--nodisplay-interp-maps',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-intvcx--2",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa',
		   '--vc-type=Int', '--nodisplay-interp-maps',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz', '--ior-type=None',
	           '--nomega-combo-topcm' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-intvcx--3",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa', '--moviefast',
		   '--vc-type=Int', '--nodisplay-interp-maps',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz', '--ior-type=None',
	           '--nomega-combo-topcm', '--vcx-outfac=5e-13' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-intvcx--4",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa', '--moviefast',
		   '--vc-type=Int', '--nodisplay-interp-maps',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz', '--ior-type=None',
	           '--nomega-combo-topcm', '--vcx-outfac=5e-13',
		   '--num-orient=6', '--num-directions=6' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-envvcx--1",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa',
		   '--vc-type=Env', '--nodisplay-interp-maps',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-envvcx--2",
	 args  => ['-KZ', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--textlog=test.txt', '--nodisplay-foa', '--moviefast',
		   '--vc-type=Env', '--nodisplay-interp-maps',
		   '--in=../inputs/ezin#.png', '--input-frames=@6Hz',
		   '--out=png', '--output-frames=@6Hz', '--evc-multithreaded',
		   '--evc-outfac=1e-17' ],
	 files => ['test.txt', 'T000000.png', 'T000001.png',
		   'T000002.png', 'T000003.png', 'T000004.png'],
     },

     {
	 name  => "ez-retina--1",
	 args  => ['-ZT', '--boring-delay=FOREVER', '--boring-sm-mv=0.0',
		   '--nodisplay-eye', '--nodisplay-eye-traj',
		   '--textlog=test.txt', '--output-frames=0-4@250',
		   '--foveate-input-depth=5',
		   '--in=raster:testpic001.pnm', '-+', '--out=ppm:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-saliency-map--5",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=0-5@30', '--save-salmap',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=pnm:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm',
		   'SM000004.pfm', 'SM000005.pfm'],
     },

     {
	 name  => "ez-saliency-map--6",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=0-5@30', '--save-salmap',
		   '--maxnorm-type=Maxnorm', '--in=raster:../inputs/ezframe#.pnm',
		   '--out=raster:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm',
		   'SM000004.pfm', 'SM000005.pfm'],
     },

     {
	 name  => "ez-saliency-map--7",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=0-5@30', '--save-salmap',
		   '--nouse-older-version',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm',
		   'SM000004.pfm', 'SM000005.pfm'],
     },

     {
	 name  => "ez-saliency-map--8",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=0-5@30', '--save-salmap',
		   '--sm-type=StdOptim',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm',
		   'SM000004.pfm', 'SM000005.pfm'],
     },

     {
	 name  => "ez-saliency-map--9",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=0-5@30', '--save-salmap',
		   '--wta-type=StdOptim',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm',
		   'SM000004.pfm', 'SM000005.pfm'],
     },

     {
	 name  => "ez-saliency-map--a",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=0-5@30', '--save-salmap',
		   '--sm-type=StdOptim --wta-type=StdOptim',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm',
		   'SM000004.pfm', 'SM000005.pfm'],
     },

     {
	 name  => "ez-saliency-trm--1",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=0-5@30', '--trm-type=KillStatic',
		   '--save-trm', '--save-trm-x-combo',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm',
		   'T000004.pnm', 'T000005.pnm',
		   'TRM000000.pfm', 'TRM000001.pfm',
		   'TRM000002.pfm', 'TRM000003.pfm',
		   'TRM000004.pfm', 'TRM000005.pfm',
		   'TRM-SB000000.pnm', 'TRM-SB000001.pnm',
		   'TRM-SB000002.pnm', 'TRM-SB000003.pnm',
		   'TRM-SB000004.pnm', 'TRM-SB000005.pnm'],
     },

     {
	 name  => "ez-variable-rate--1",
	 args  => ['-Z', '--textlog=test.txt',
		   '--input-frames=0-5@../inputs/variablerate.fl',
		   '--output-frames=0-5@30',
		   '--save-salmap', '--maxnorm-type=Maxnorm',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt', 'SM000000.pfm', 'SM000001.pfm',
		   'SM000002.pfm', 'SM000003.pfm',
		   'SM000004.pfm', 'SM000005.pfm'],
     },

     {
	 name  => "ez-save-everything--1",
	 args  => ['-Z', '--textlog=test.txt', '--input-frames=0-5@30',
		   '--output-frames=100-103@40',
		   '--trm-type=Std', '--shape-estim-mode=ConspicuityMap',
		   '--save-retina-input', '--save-retina-output',
		   '--save-retina-pyr', '--save-objmask', '--save-agm',
		   '--save-trm', '--save-salmap', '--save-cumsalmap',
		   '--save-wta', '--save-vcx-output',
		   '--save-featurecomb-maps', '--save-conspic-maps',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:',
		   '--out=info:info.txt'],
	 files => ['test.txt', 'info.txt',
		   'AGM000100.pfm', 'AGM000101.pfm', 'AGM000102.pfm', 'AGM000103.pfm',
		   'COcolor-000100.pnm', 'COcolor-000101.pnm', 'COcolor-000102.pnm', 'COcolor-000103.pnm',
		   'COmotion-000100.pnm', 'COmotion-000101.pnm', 'COmotion-000102.pnm', 'COmotion-000103.pnm',
		   'COorientation-000100.pnm', 'COorientation-000101.pnm', 'COorientation-000102.pnm', 'COorientation-000103.pnm',
		   'CUMSM000100.pfm', 'CUMSM000101.pfm', 'CUMSM000102.pfm', 'CUMSM000103.pfm',
		   'OBJ000100.pnm', 'OBJ000101.pnm', 'OBJ000102.pnm', 'OBJ000103.pnm',
		   'RETIN-000100.pnm', 'RETIN-000101.pnm', 'RETIN-000102.pnm', 'RETIN-000103.pnm',
		   'RETOUT-000100.pnm', 'RETOUT-000101.pnm', 'RETOUT-000102.pnm', 'RETOUT-000103.pnm',
		   'SM000100.pfm', 'SM000101.pfm', 'SM000102.pfm', 'SM000103.pfm',
		   'SOby-000100.pnm', 'SOby-000101.pnm', 'SOby-000102.pnm', 'SOby-000103.pnm',
		   'SOdir_0-000100.pnm', 'SOdir_0-000101.pnm', 'SOdir_0-000102.pnm', 'SOdir_0-000103.pnm',
		   'SOdir_1-000100.pnm', 'SOdir_1-000101.pnm', 'SOdir_1-000102.pnm', 'SOdir_1-000103.pnm',
		   'SOdir_2-000100.pnm', 'SOdir_2-000101.pnm', 'SOdir_2-000102.pnm', 'SOdir_2-000103.pnm',
		   'SOdir_3-000100.pnm', 'SOdir_3-000101.pnm', 'SOdir_3-000102.pnm', 'SOdir_3-000103.pnm',
		   'SOflicker-000100.pnm', 'SOflicker-000101.pnm', 'SOflicker-000102.pnm', 'SOflicker-000103.pnm',
		   'SOintensity-000100.pnm', 'SOintensity-000101.pnm', 'SOintensity-000102.pnm', 'SOintensity-000103.pnm',
		   'SOori_0-000100.pnm', 'SOori_0-000101.pnm', 'SOori_0-000102.pnm', 'SOori_0-000103.pnm',
		   'SOori_1-000100.pnm', 'SOori_1-000101.pnm', 'SOori_1-000102.pnm', 'SOori_1-000103.pnm',
		   'SOori_2-000100.pnm', 'SOori_2-000101.pnm', 'SOori_2-000102.pnm', 'SOori_2-000103.pnm',
		   'SOori_3-000100.pnm', 'SOori_3-000101.pnm', 'SOori_3-000102.pnm', 'SOori_3-000103.pnm',
		   'SOrg-000100.pnm', 'SOrg-000101.pnm', 'SOrg-000102.pnm', 'SOrg-000103.pnm',
		   'TRM000100.pfm', 'TRM000101.pfm', 'TRM000102.pfm', 'TRM000103.pfm',
		   'VCO000100.pfm', 'VCO000101.pfm', 'VCO000102.pfm', 'VCO000103.pfm',
		   'WTA000100.pfm', 'WTA000101.pfm', 'WTA000102.pfm', 'WTA000103.pfm'],
     },

     {
	 name  => "ez-junction-channels--2",
	 args  => ['-Z', '--textlog=test.txt', '--vc-chans=OLTXE',
		   '--input-frames=0-5@35', '--output-frames=100-103@41',
		   '--save-featurecomb-maps', '--save-conspic-maps',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt',
		   'COEndPoint-000100.pnm', 'COEndPoint-000101.pnm', 'COEndPoint-000102.pnm', 'COEndPoint-000103.pnm',
		   'COLJunction-000100.pnm', 'COLJunction-000101.pnm', 'COLJunction-000102.pnm', 'COLJunction-000103.pnm',
		   'COTJunction-000100.pnm', 'COTJunction-000101.pnm', 'COTJunction-000102.pnm', 'COTJunction-000103.pnm',
		   'COXJunction-000100.pnm', 'COXJunction-000101.pnm', 'COXJunction-000102.pnm', 'COXJunction-000103.pnm',
		   'COorientation-000100.pnm', 'COorientation-000101.pnm', 'COorientation-000102.pnm', 'COorientation-000103.pnm',
		   'SOjunction_00000010-000100.pnm', 'SOjunction_00000010-000101.pnm', 'SOjunction_00000010-000102.pnm', 'SOjunction_00000010-000103.pnm',
		   'SOjunction_00001000-000100.pnm', 'SOjunction_00001000-000101.pnm', 'SOjunction_00001000-000102.pnm', 'SOjunction_00001000-000103.pnm',
		   'SOjunction_00001010-000100.pnm', 'SOjunction_00001010-000101.pnm', 'SOjunction_00001010-000102.pnm', 'SOjunction_00001010-000103.pnm',
		   'SOjunction_00100000-000100.pnm', 'SOjunction_00100000-000101.pnm', 'SOjunction_00100000-000102.pnm', 'SOjunction_00100000-000103.pnm',
		   'SOjunction_00101000-000100.pnm', 'SOjunction_00101000-000101.pnm', 'SOjunction_00101000-000102.pnm', 'SOjunction_00101000-000103.pnm',
		   'SOjunction_00101010-000100.pnm', 'SOjunction_00101010-000101.pnm', 'SOjunction_00101010-000102.pnm', 'SOjunction_00101010-000103.pnm',
		   'SOjunction_10000000-000100.pnm', 'SOjunction_10000000-000101.pnm', 'SOjunction_10000000-000102.pnm', 'SOjunction_10000000-000103.pnm',
		   'SOjunction_10000010-000100.pnm', 'SOjunction_10000010-000101.pnm', 'SOjunction_10000010-000102.pnm', 'SOjunction_10000010-000103.pnm',
		   'SOjunction_10001010-000100.pnm', 'SOjunction_10001010-000101.pnm', 'SOjunction_10001010-000102.pnm', 'SOjunction_10001010-000103.pnm',
		   'SOjunction_10100000-000100.pnm', 'SOjunction_10100000-000101.pnm', 'SOjunction_10100000-000102.pnm', 'SOjunction_10100000-000103.pnm',
		   'SOjunction_10100010-000100.pnm', 'SOjunction_10100010-000101.pnm', 'SOjunction_10100010-000102.pnm', 'SOjunction_10100010-000103.pnm',
		   'SOjunction_10101000-000100.pnm', 'SOjunction_10101000-000101.pnm', 'SOjunction_10101000-000102.pnm', 'SOjunction_10101000-000103.pnm',
		   'SOjunction_10101010-000100.pnm', 'SOjunction_10101010-000101.pnm', 'SOjunction_10101010-000102.pnm', 'SOjunction_10101010-000103.pnm',
		   'SOori_0-000100.pnm', 'SOori_0-000101.pnm', 'SOori_0-000102.pnm', 'SOori_0-000103.pnm',
		   'SOori_1-000100.pnm', 'SOori_1-000101.pnm', 'SOori_1-000102.pnm', 'SOori_1-000103.pnm',
		   'SOori_2-000100.pnm', 'SOori_2-000101.pnm', 'SOori_2-000102.pnm', 'SOori_2-000103.pnm',
		   'SOori_3-000100.pnm', 'SOori_3-000101.pnm', 'SOori_3-000102.pnm', 'SOori_3-000103.pnm'],
     },

     {
	 name  => "ez-extra-channels--1",
	 args  => ['-Z', '--textlog=test.txt', '--vc-chans=SGNR',
		   '--input-frames=0-5@25', '--output-frames=100-102@41',
		   '--save-featurecomb-maps', '--save-conspic-maps',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt',
		   'COIntensityBand-000100.pnm', 'COIntensityBand-000101.pnm', 'COIntensityBand-000102.pnm',
		   'COMultiColorBand-000100.pnm', 'COMultiColorBand-000101.pnm', 'COMultiColorBand-000102.pnm',
		   'COcolor-000100.pnm', 'COcolor-000101.pnm', 'COcolor-000102.pnm',
		   'COcomposite-color-000100.pnm', 'COcomposite-color-000101.pnm', 'COcomposite-color-000102.pnm',
		   'COsingle-opponent-color-000100.pnm', 'COsingle-opponent-color-000101.pnm', 'COsingle-opponent-color-000102.pnm',
		   'SOb-000100.pnm', 'SOb-000101.pnm', 'SOb-000102.pnm',
		   'SOb-y-000100.pnm', 'SOb-y-000101.pnm', 'SOb-y-000102.pnm',
		   'SOby-000100.pnm', 'SOby-000101.pnm', 'SOby-000102.pnm',
		   'SOcband_0-000100.pnm', 'SOcband_0-000101.pnm', 'SOcband_0-000102.pnm',
		   'SOcband_1-000100.pnm', 'SOcband_1-000101.pnm', 'SOcband_1-000102.pnm',
		   'SOcband_2-000100.pnm', 'SOcband_2-000101.pnm', 'SOcband_2-000102.pnm',
		   'SOcband_3-000100.pnm', 'SOcband_3-000101.pnm', 'SOcband_3-000102.pnm',
		   'SOcband_4-000100.pnm', 'SOcband_4-000101.pnm', 'SOcband_4-000102.pnm',
		   'SOcband_5-000100.pnm', 'SOcband_5-000101.pnm', 'SOcband_5-000102.pnm',
		   'SOg-000100.pnm', 'SOg-000101.pnm', 'SOg-000102.pnm',
		   'SOg-r-000100.pnm', 'SOg-r-000101.pnm', 'SOg-r-000102.pnm',
		   'SOiband_0-000100.pnm', 'SOiband_0-000101.pnm', 'SOiband_0-000102.pnm',
		   'SOiband_1-000100.pnm', 'SOiband_1-000101.pnm', 'SOiband_1-000102.pnm',
		   'SOiband_2-000100.pnm', 'SOiband_2-000101.pnm', 'SOiband_2-000102.pnm',
		   'SOr-000100.pnm', 'SOr-000101.pnm', 'SOr-000102.pnm',
		   'SOr-g-000100.pnm', 'SOr-g-000101.pnm', 'SOr-g-000102.pnm',
		   'SOrg-000100.pnm', 'SOrg-000101.pnm', 'SOrg-000102.pnm',
		   'SOy-000100.pnm', 'SOy-000101.pnm', 'SOy-000102.pnm',
		   'SOy-b-000100.pnm', 'SOy-b-000101.pnm', 'SOy-b-000102.pnm',
		   'SOPedestrian-000100.pnm', 'SOPedestrian-000101.pnm', 'SOPedestrian-000102.pnm'],
	 },

     {
	 name  => "ez-dkl-channel--1",
	 args  => ['-Z', '--textlog=test.txt', '--vc-chans=J',
		   '--input-frames=0-5@25', '--output-frames=100-102@41',
		   '--save-featurecomb-maps', '--save-conspic-maps', '--save-raw-maps',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt',
		   'CODKLcolor-000100.pnm', 'CODKLcolor-000101.pnm',
		   'CODKLcolor-000102.pnm', 'SOdcolor-000100.pnm',
		   'SOdcolor-000101.pnm', 'SOdcolor-000102.pnm',
		   'SOkcolor-000100.pnm', 'SOkcolor-000101.pnm',
		   'SOkcolor-000102.pnm', 'SOlcolor-000100.pnm',
		   'SOlcolor-000101.pnm', 'SOlcolor-000102.pnm',
		   'SRdcolor-0-000100.pnm', 'SRdcolor-0-000101.pnm',
		   'SRdcolor-0-000102.pnm', 'SRdcolor-1-000100.pnm',
		   'SRdcolor-1-000101.pnm', 'SRdcolor-1-000102.pnm',
		   'SRdcolor-2-000100.pnm', 'SRdcolor-2-000101.pnm',
		   'SRdcolor-2-000102.pnm', 'SRdcolor-3-000100.pnm',
		   'SRdcolor-3-000101.pnm', 'SRdcolor-3-000102.pnm',
		   'SRdcolor-4-000100.pnm', 'SRdcolor-4-000101.pnm',
		   'SRdcolor-4-000102.pnm', 'SRdcolor-5-000100.pnm',
		   'SRdcolor-5-000101.pnm', 'SRdcolor-5-000102.pnm',
		   'SRdcolor-6-000100.pnm', 'SRdcolor-6-000101.pnm',
		   'SRdcolor-6-000102.pnm', 'SRdcolor-7-000100.pnm',
		   'SRdcolor-7-000101.pnm', 'SRdcolor-7-000102.pnm',
		   'SRdcolor-8-000100.pnm', 'SRdcolor-8-000101.pnm',
		   'SRdcolor-8-000102.pnm', 'SRkcolor-0-000100.pnm',
		   'SRkcolor-0-000101.pnm', 'SRkcolor-0-000102.pnm',
		   'SRkcolor-1-000100.pnm', 'SRkcolor-1-000101.pnm',
		   'SRkcolor-1-000102.pnm', 'SRkcolor-2-000100.pnm',
		   'SRkcolor-2-000101.pnm', 'SRkcolor-2-000102.pnm',
		   'SRkcolor-3-000100.pnm', 'SRkcolor-3-000101.pnm',
		   'SRkcolor-3-000102.pnm', 'SRkcolor-4-000100.pnm',
		   'SRkcolor-4-000101.pnm', 'SRkcolor-4-000102.pnm',
		   'SRkcolor-5-000100.pnm', 'SRkcolor-5-000101.pnm',
		   'SRkcolor-5-000102.pnm', 'SRkcolor-6-000100.pnm',
		   'SRkcolor-6-000101.pnm', 'SRkcolor-6-000102.pnm',
		   'SRkcolor-7-000100.pnm', 'SRkcolor-7-000101.pnm',
		   'SRkcolor-7-000102.pnm', 'SRkcolor-8-000100.pnm',
		   'SRkcolor-8-000101.pnm', 'SRkcolor-8-000102.pnm',
		   'SRlcolor-0-000100.pnm', 'SRlcolor-0-000101.pnm',
		   'SRlcolor-0-000102.pnm', 'SRlcolor-1-000100.pnm',
		   'SRlcolor-1-000101.pnm', 'SRlcolor-1-000102.pnm',
		   'SRlcolor-2-000100.pnm', 'SRlcolor-2-000101.pnm',
		   'SRlcolor-2-000102.pnm', 'SRlcolor-3-000100.pnm',
		   'SRlcolor-3-000101.pnm', 'SRlcolor-3-000102.pnm',
		   'SRlcolor-4-000100.pnm', 'SRlcolor-4-000101.pnm',
		   'SRlcolor-4-000102.pnm', 'SRlcolor-5-000100.pnm',
		   'SRlcolor-5-000101.pnm', 'SRlcolor-5-000102.pnm',
		   'SRlcolor-6-000100.pnm', 'SRlcolor-6-000101.pnm',
		   'SRlcolor-6-000102.pnm', 'SRlcolor-7-000100.pnm',
		   'SRlcolor-7-000101.pnm', 'SRlcolor-7-000102.pnm',
		   'SRlcolor-8-000100.pnm', 'SRlcolor-8-000101.pnm',
		   'SRlcolor-8-000102.pnm'],
     },

     {
	 name  => "ez-target-mask--1",
	 args  => ['-ZT', '--boring-delay=max', '--nodisplay-eye',
		   '--nodisplay-eye-traj', '--textlog=test.txt',
		   '--output-frames=@EVENT', '-+',
		   '--in=raster:testpic001.pnm', '--out=raster',
		   '--target-mask=../inputs/targetmask001.png'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm',
		   'T000002.pnm', 'T000003.pnm', 'T000004.pnm'],
     },

     {
	 name  => "ez-eye-trace--1",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/ezframe#.pnm', '--input-frames=@6Hz',
		   '--out=pnm', '--output-frames=@3Hz',
		   '--vc-type=None', '--sm-type=None', '--trm-type=None',
		   '--wta-type=None', '--shape-estim-mode=None',
		   '--sv-type=EyeMvt', '--save-trajectory',
		   '--fovea-radius=32', '--maxcache-size=0',
		   '--eyetrack-data=../inputs/eyetrace.eye'],
	 files => ['T000000.pnm', 'T000001.pnm', 'T000002.pnm'],
     },

     {
	 name  => "ez-eye-trace--2",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/ezframe#.pnm', '--input-frames=@6Hz',
		   '--out=pnm', '--output-frames=@3Hz',
		   '--shape-estim-mode=None', '--wta-type=None',
		   '--ehc-type=EyeTrack', '--sv-type=Compress', '--mega-combo',
		   '--maxnorm-type=Maxnorm', '--nodisplay-interp-maps',
		   '--eyetrack-data=../inputs/eyetrace.eye',
		   '--svem-num-random=0',
		   '--svem-out-fname=eyeout.txt' ],
	 files => ['T000000.pnm', 'T000001.pnm', 'T000002.pnm', 'eyeout.txt'],
     },

     {
	 name  => "ez-eye-trace--3",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/ezframe#.pnm', '--input-frames=@6Hz',
		   '--out=pnm', '--output-frames=@3Hz',
		   '--shape-estim-mode=None', '--wta-type=None',
		   '--ehc-type=EyeTrack', '--sv-type=Compress', '--mega-combo',
		   '--maxnorm-type=Maxnorm', '--nodisplay-interp-maps',
		   '--eyetrack-data=../inputs/eyetrace.eye',
		   '--svem-out-fname=eyeout.txt', '--num-foveas=3' ],
	 files => ['T000000.pnm', 'T000001.pnm', 'T000002.pnm'],
     },

     {
	 name  => "ez-eye-trace--4",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/ezin#.png', '--input-frames=@1.5Hz',
		   '--rescale-input=640x480', '--out=pnm', '--output-frames=@1.5Hz',
		   '--eyecompare', '--maxnorm-type=Maxnorm', '--nodisplay-interp-maps',
		   '--eyetrack-data=../inputs/eyetrace2.e-ceyeS',
		   '--svem-out-fname=eyeout.txt' ],
	 files => ['T000000.pnm', 'T000001.pnm', 'T000002.pnm', 'T000003.pnm', 'T000004.pnm',],
     },

#    single video, single eyetrace, special fields
     {
	 name  => "ez-eyesal-output--1",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/hdclip1.mpg', '--input-frames=@30Hz',
		   '--out=null', '--movie', '--eyecompare',
		   '--sv-type=EyeMvt', '--fovea-radius=35',
		   '--svem-num-random=0', '--vc-chans=I',
		   '--vcx-load-out-from=../inputs/hdclip1.mgz',
		   '--eyetrack-data=../inputs/subj1.e-ceyeS',
		   '--svem-out-fname=subj1.eyesal'],
	 files => ['subj1.eyesal'],
     },
#    single video, single eyetrace, full un-marked fields
{
	 name  => "ez-eyesal-output--2",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/hdclip1.mpg', '--input-frames=@30Hz',
		   '--out=null', '--movie', '--eyecompare',
		   '--sv-type=EyeMvt', '--fovea-radius=35',
		   '--svem-num-random=0', '--vc-chans=I',
		   '--vcx-load-out-from=../inputs/hdclip1.mgz',
		   '--eyetrack-data=../inputs/subj1-nohead.e-ceyeS',
		   '--svem-out-fname=subj1-nohead.eyesal'],
	 files => ['subj1-nohead.eyesal'],
     },
#    single video, multiple eyetrace, special fields
{
	 name  => "ez-eyesal-output--3",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/hdclip1.mpg', '--input-frames=@30Hz',
		   '--out=null', '--movie', '--eyecompare',
		   '--sv-type=EyeMvt', '--fovea-radius=35',
		   '--svem-num-random=0', '--vc-chans=I',
		   '--vcx-load-out-from=../inputs/hdclip1.mgz',
		   '--eyetrack-data=../inputs/subj1.e-ceyeS,../inputs/subj2.e-ceyeS,../inputs/subj3.e-ceyeS',
		   '--svem-out-fname=subj123.eyesal'],
	 files => ['subj123.eyesal'],
     },
#    single video, regions, single eyetrace, unmarked fields
{
	 name  => "ez-eyesal-output--4",
	 args  => ['--nouse-fpe', '--nouse-random',
		   '--in=../inputs/hdclip1.mpg', '--input-frames=@30Hz',
		   '--out=null', '--movie', '--eyecompare',
		   '--sv-type=EyeRegion', '--fovea-radius=35',
		   '--svem-num-random=0', '--vc-chans=I',
		   '--vcx-load-out-from=../inputs/hdclip1.mgz',
		   '--eyetrack-data=../inputs/subj1.e-ceyeS',
		   '--xml-file=../inputs/hdclip1.xml',
		   '--svem-out-fname=subj1_regions.eyesal'],
	 files => ['subj1_regions.eyesal'],
     },
#     {
#	 name  => "ez-eye-trace--5",
#	 args  => ['--nouse-fpe', '--nouse-random',
#		   '--in=../inputs/ezin#.png', '--input-frames=@1.5Hz',
#		   '--rescale-input=640x480', '--out=pnm', '--output-frames=@1.5Hz',
#		   '--eyecompare', '--maxnorm-type=Maxnorm', '--nodisplay-interp-maps',
#		   '--eyetrack-data=../inputs/eyetrace2.e-ceyeS',
#		   '--svem-out-fname=eyeout.txt', '--shift-input' ],
#	 files => ['T000000.pnm', 'T000001.pnm', 'T000002.pnm', 'T000003.pnm', 'T000004.pnm',],
#     },

     {
	 name  => "ez-surprise-channels--1",
	 args  => ['-Z', '--textlog=test.txt', '--vc-type=Surp',
		   '--maxnorm-type=Surprise',
		   '--nouse-older-version',
		   '--chanout-max=0.0',
		   '--gabor-intens=20.0',
		   '--direction-sqrt',
		   '--surprise-qlen=5',
		   '--surprise-updfac=0.7',
		   '--surprise-neighupdfac=0.7',
		   '--surprise-probe=-1,-1',
		   '--surprise-slfac=1.0',
		   '--surprise-ssfac=0.05',
		   '--surprise-logged=false',
		   '--surprise-type=Poisson',
		   '--input-frames=0-2@30', '--output-frames=0-2@30',
		   '--save-featurecomb-maps', '--save-conspic-maps',
		   '--in=raster:../inputs/ezframe#.pnm', '--out=raster:'],
	 files => ['test.txt',
		   'SOby-000000.pnm', 'SOby-000001.pnm', 'SOby-000002.pnm',
		   'SOdir_0-000000.pnm', 'SOdir_0-000001.pnm', 'SOdir_0-000002.pnm',
		   'SOdir_1-000000.pnm', 'SOdir_1-000001.pnm', 'SOdir_1-000002.pnm',
		   'SOdir_2-000000.pnm', 'SOdir_2-000001.pnm', 'SOdir_2-000002.pnm',
		   'SOdir_3-000000.pnm', 'SOdir_3-000001.pnm', 'SOdir_3-000002.pnm',
		   'SOflicker-000000.pnm', 'SOflicker-000001.pnm', 'SOflicker-000002.pnm',
		   'SOintensity-000000.pnm', 'SOintensity-000001.pnm', 'SOintensity-000002.pnm',
		   'SOori_0-000000.pnm', 'SOori_0-000001.pnm', 'SOori_0-000002.pnm',
		   'SOori_1-000000.pnm', 'SOori_1-000001.pnm', 'SOori_1-000002.pnm',
		   'SOori_2-000000.pnm', 'SOori_2-000001.pnm', 'SOori_2-000002.pnm',
		   'SOori_3-000000.pnm', 'SOori_3-000001.pnm', 'SOori_3-000002.pnm',
		   'SOrg-000000.pnm', 'SOrg-000001.pnm', 'SOrg-000002.pnm',
		   'COcolor-000000.pnm', 'COcolor-000001.pnm', 'COcolor-000002.pnm',
		   'COmotion-000000.pnm', 'COmotion-000001.pnm', 'COmotion-000002.pnm',
		   'COorientation-000000.pnm', 'COorientation-000001.pnm', 'COorientation-000002.pnm',
		   ],
     },

     {
	 name  => "ez-spectral-residual--1",
	 args  => ['-Z', '--in=../inputs/testpic001.pnm', '--out=pnm',
		   '--vc-type=SpectralResidual', '--maxnorm-type=Ignore',
		   '--srs-lowpass-filter-width=8', '--trm-type=None',
		   '--wta-type=None', '--sm-type=None',
		   '--sv-type=None',
		   '--save-srs-output', '--save-srs-extra-output',
		   '--srs-resize=/4', '--srs-output-blur=3.5',
		   '--srs-spectral-blur=7', '--textlog=test.txt'],
	 files => ['test.txt',
		   'SRS-000000.pfm',
		   'SRS-rescaled-input000000.pfm',
		   'SRS-logmagnitude000000.pfm',
		   'SRS-phase000000.pfm',
		   'SRS-spectral-residual000000.pfm',
		   'SRS-proto-saliency000000.pfm'],
     },

     {
	 name  => "ez-spectral-residual--2",
	 args  => ['-Z', '--in=../inputs/testpic001.pnm', '--out=pnm',
		   '--vc-type=SpectralResidual', '--maxnorm-type=Ignore',
		   '--trm-type=None', '--wta-type=None', '--sm-type=None',
		   '--sv-type=None',
		   '--save-srs-output', '--save-srs-extra-output',
		   '--srs-resize=noop', '--srs-output-blur=3.5',
		   '--srs-spectral-blur=7', '--textlog=test.txt'],
	 files => ['test.txt',
		   'SRS-000000.pfm',
		   'SRS-rescaled-input000000.pfm',
		   'SRS-logmagnitude000000.pfm',
		   'SRS-phase000000.pfm',
		   'SRS-spectral-residual000000.pfm',
		   'SRS-proto-saliency000000.pfm'],
     },

     {
	 name  => "ez-multi-spectral-residual--1",
	 args  => ['-Z', '--in=../inputs/testpic001.pnm', '--out=pnm',
		   '--vc-type=MultiSpectralResidual',
		   '--maxnorm-type=None',
		   '--trm-type=None', '--wta-type=None', '--sm-type=None',
		   '--sv-type=None',
		   '--multi-srs-sizes=16x16,32x32,64x64,128x128',
		   '--srs-output-blur-factor=0.055',
		   '--srs-spectral-blur=1',
		   '--srs-attenuation-width=0.0625',
		   '--srs-lowpass-filter-width=2',
		   '--map-combine-type=Max',
		   '--save-srs-output', '--save-conspic-maps',
		   '--textlog=test.txt'],
	 files => ['test.txt',
		   'COMultiSpectralResidualChannel-000000.pnm',
		   'srs_0-000000.pfm',
		   'srs_1-000000.pfm',
		   'srs_2-000000.pfm',
		   'srs_3-000000.pfm'],
     },

     {
	 name  => "ez-multi-spectral-residual--2",
	 args  => ['-Z', '--in=../inputs/testpic001.pnm', '--out=pnm',
		   '--vc-type=MultiSpectralResidual',
		   '--maxnorm-type=None',
		   '--trm-type=None', '--wta-type=None', '--sm-type=None',
		   '--sv-type=None',
		   '--multi-srs-sizes=17x17,25x19,31x33,50x52',
		   '--srs-output-blur-factor=0.055',
		   '--srs-spectral-blur=1',
		   '--srs-attenuation-width=0.0625',
		   '--map-combine-type=Max',
		   '--save-srs-output', '--save-srs-extra-output',
		   '--save-conspic-maps',
		   '--textlog=test.txt'],
	 files => ['test.txt',
		   'COMultiSpectralResidualChannel-000000.pnm',
		   'srs_0-000000.pfm',
		   'srs_0-rescaled-input000000.pfm',
		   'srs_0-logmagnitude000000.pfm',
		   'srs_0-phase000000.pfm',
		   'srs_0-spectral-residual000000.pfm',
		   'srs_0-proto-saliency000000.pfm',
		   'srs_1-000000.pfm',
		   'srs_1-rescaled-input000000.pfm',
		   'srs_1-logmagnitude000000.pfm',
		   'srs_1-phase000000.pfm',
		   'srs_1-spectral-residual000000.pfm',
		   'srs_1-proto-saliency000000.pfm',
		   'srs_2-000000.pfm',
		   'srs_2-rescaled-input000000.pfm',
		   'srs_2-logmagnitude000000.pfm',
		   'srs_2-phase000000.pfm',
		   'srs_2-spectral-residual000000.pfm',
		   'srs_2-proto-saliency000000.pfm',
		   'srs_3-000000.pfm',
		   'srs_3-rescaled-input000000.pfm',
		   'srs_3-logmagnitude000000.pfm',
		   'srs_3-phase000000.pfm',
		   'srs_3-spectral-residual000000.pfm',
		   'srs_3-proto-saliency000000.pfm'],
     },

     {
	 name  => "ez-multi-spectral-residual--3",
	 args  => ['-Z', '--in=../inputs/testpic001.pnm', '--out=pnm',
		   '--vc-type=MultiSpectralResidual',
		   '--maxnorm-type=None',
		   '--trm-type=None', '--wta-type=None', '--sm-type=None',
		   '--sv-type=None',
		   '--multi-srs-sizes=/8,/16,/4,noop',
		   '--srs-output-blur-factor=0.055',
		   '--srs-spectral-blur=1',
		   '--srs-attenuation-width=0.0625',
		   '--map-combine-type=Max',
		   '--save-srs-output', '--save-conspic-maps',
		   '--textlog=test.txt'],
	 files => ['test.txt',
		   'COMultiSpectralResidualChannel-000000.pnm',
		   'srs_0-000000.pfm',
		   'srs_1-000000.pfm',
		   'srs_2-000000.pfm',
		   'srs_3-000000.pfm'],
     },

     {
	 name  => "ez-optigains--1",
	 args  => ['-Z', '--train-optigains', '--in=../inputs/berry1.xml',
		   '--stsd-filename=berry1.stsd'],
	 files => ['berry1.stsd'],
     },

     {
	 name  => "ez-optigains--2",
	 args  => ['-Z', '--train-optigains', '--in=../inputs/berry2.xml',
		   '--stsd-filename=berry2.stsd'],
	 files => ['berry2.stsd'],
     },

     {
	 name  => "ez-optigains--3",
	 args  => ['-Z', '--guided-search', '--in=../inputs/berry3.xml',
		   '--gains-filename=../ref/combine-optigains--1--berry12.pmap',
		   '-K', '--out=pnm', '--output-frames=0-2@EVENT',
		   '--nodisplay-interp-maps', '-+', '--textlog=test.txt'],
	 files => ['test.txt', 'T000000.pnm', 'T000001.pnm', 'T000002.pnm'],
     },

     # this provides reference outputs which we then use in test_simplesaliency_blackbox.pl to test whether
     # simple-saliency generates exactly the same outputs:
     {
	 name  => "ez-simplesaliency--1",
	 args  => ['-XZ', '--sm-type=None --wta-type=None',
		   '--textlog=test.txt', '--nodisplay-foa --nodisplay-time',
		   '--nodisplay-interp-maps', '--maxnorm-type=Maxnorm',
		   '--in=../inputs/ezin#.png', '--input-frames=@30ms',
		   '--out=png', '--output-frames=@30ms', '--out=pfm',
		   '--nodisplay-interp-maps', '--display-map=VCO', '--save-vcx-output' ],
	 files => ['test.txt',
		   'T000000.png', 'T000001.png', 'T000002.png', 'T000003.png', 'T000004.png',
		   'VCO000000.png', 'VCO000001.png', 'VCO000002.png', 'VCO000003.png', 'VCO000004.png',
		   'T000000.pfm', 'T000001.pfm', 'T000002.pfm', 'T000003.pfm', 'T000004.pfm',
		   'VCO000000.pfm', 'VCO000001.pfm', 'VCO000002.pfm', 'VCO000003.pfm', 'VCO000004.pfm' ],
     },


    );


# Run the black box tests; note that the default executable can be
# overridden from the command-line with "--executable"

blackbox::run("$invt_config::exec_prefix/bin/ezvision", @tests);
